/*
 *  UFTP - UDP based FTP with multicast
 *
 *  Copyright (C) 2001-2008   Dennis A. Bush, Jr.   bush@tcnj.edu
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/timeb.h>
#include <stdarg.h>
#include <errno.h>

#if ( defined WIN16 || defined WIN32 || defined WIN64 )

#define WINDOWS
#include <winsock2.h>
#include <ws2tcpip.h>
#include <winbase.h>
#include <io.h>
#include <process.h>
#include <winnt.h>
#include "win_func.h"
#include "winsock_err.h"

typedef unsigned __int16 uint16_t;
typedef __int16 int16_t;
typedef unsigned __int32 uint32_t;
typedef __int32 int32_t;
typedef unsigned __int64 uint64_t;
typedef __int64 int64_t;

#define link(src,dest) rename(src,dest)
#define bzero(buf,size) memset(buf,0,size)
#define usleep(t) Sleep((t)/1000)
#define socketerror(str) fprintf(stderr,"%s: (%d) %s\n",str,WSAGetLastError(),GetWinsockErrString(WSAGetLastError()))
#define isfullpath(str) (str[1]==':')

typedef int64_t f_offset_t;
typedef struct __stat64 stat_struct;
#define stat_func(p,s) _stat64(p,s)
#define lseek_func(f,o,l) _lseeki64(f,o,l)
#define snprintf _snprintf

typedef LARGE_INTEGER timestamp;
#define gettimestamp(t) QueryPerformanceCounter(&t)

int32_t diff_sec(timestamp t2, timestamp t1)
{ 
  timestamp f;
  
  QueryPerformanceFrequency(&f);
  return (int32_t)((t2.QuadPart-t1.QuadPart)/f.QuadPart);
}

int32_t diff_usec(timestamp t2, timestamp t1)
{ 
  timestamp f;
  
  QueryPerformanceFrequency(&f);
  return (int32_t)(1000000*(t2.QuadPart-t1.QuadPart)/f.QuadPart);
}

timestamp add_msec_interval(timestamp t, int interval)
{
  timestamp f,new_interval,new_time;
  
  QueryPerformanceFrequency(&f);
  new_interval.QuadPart=(LONG64)interval*f.QuadPart/1000;
  new_time.QuadPart=t.QuadPart+new_interval.QuadPart;
  return new_time;
}

timestamp add_sec_interval(timestamp t, int interval)
{
  timestamp f,new_interval,new_time;
  
  QueryPerformanceFrequency(&f);
  new_interval.QuadPart=(LONG64)interval*f.QuadPart;
  new_time.QuadPart=t.QuadPart+new_interval.QuadPart;
  return new_time;
}

int cmptimestamp(timestamp t1, timestamp t2)
{
  if (t1.QuadPart>t2.QuadPart) {
    return 1;
  } else if (t1.QuadPart<t2.QuadPart) {
    return -1;
  } else {
    return 0;
  }
}

char *printll(int64_t n)
{
  static char str[50];
  snprintf(str,sizeof(str)-1,"%I64d",n);
  str[sizeof(str)-1]='\x0';
  return &str[0];
}

#define OPENREAD (O_RDONLY|O_BINARY)
#define OPENWRITE (O_WRONLY|O_BINARY)

#define PATH_SEP '\\'
#define MICRO '\xe6'

#else  /*if WINDOWS*/

#include <netinet/in.h>
#include <sys/socket.h>
#include <poll.h>
#include <unistd.h>
#include <netdb.h>
#include <strings.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <libgen.h>

#define closesocket(s) close(s)
#define socketerror(str) perror(str)
#define isfullpath(str) (str[0]=='/')

typedef int64_t f_offset_t;
typedef struct stat stat_struct;
#define stat_func(p,s) stat(p,s)
#define lseek_func(f,o,l) lseek(f,o,l)

typedef struct timeval timestamp;
#define gettimestamp(t) gettimeofday(&t,NULL)

int32_t diff_sec(timestamp t2, timestamp t1)
{
  return t2.tv_sec-t1.tv_sec;
}

int32_t diff_usec(timestamp t2, timestamp t1)
{
  return (t2.tv_usec-t1.tv_usec)+1000000*(t2.tv_sec-t1.tv_sec);
}

timestamp add_msec_interval(timestamp t, int interval)
{
  timestamp new_time;
  
  new_time=t;
  new_time.tv_usec+=interval*1000;
  while (new_time.tv_usec>=1000000) {
    new_time.tv_usec-=1000000;
    new_time.tv_sec+=1;
  }
  return new_time;
}

timestamp add_sec_interval(timestamp t, int interval)
{
  timestamp new_time;
  
  new_time=t;
  new_time.tv_sec+=interval;
  return new_time;
}

int cmptimestamp(timestamp t1, timestamp t2)
{
  if (t1.tv_sec>t2.tv_sec) {
    return 1;
  } else if (t1.tv_sec<t2.tv_sec) {
    return -1;
  } else if (t1.tv_usec>t2.tv_usec) {
    return 1;
  } else if (t1.tv_usec<t2.tv_usec) {
    return -1;
  } else {
    return 0;
  }
}

char *printll(int64_t n)
{
  static char str[50];
  snprintf(str,sizeof(str)-1,"%lld",n);
  str[sizeof(str)-1]='\x0';
  return &str[0];
}

typedef int SOCKET;
#define INVALID_SOCKET -1
#define SOCKET_ERROR -1
#ifndef INADDR_NONE
#define INADDR_NONE -1
#endif
#define OPENREAD (O_RDONLY)
#define OPENWRITE (O_WRONLY)

#define PATH_SEP '/'
#define MICRO '\xe6'

#endif /*if WINDOWS*/

#define UFTP_ID		0x69
#define ANNOUNCE	1
#define REGISTER	2
#define REG_CONF	3
#define FILESEG		4
#define DONE		5
#define STATUS		6
#define DONE_CONF	7
#define ABORT		99

#define BLOCKSIZE 1400
#define PACKETSIZE (BLOCKSIZE+sizeof(struct uftp_h))
#define MAXDEST 100
#define MAXOPENDEST 10000
#define MAXFILENAME 100
#define MAXDIRNAME 200
#define MAXPATHNAME 300

#define VERSION "UFTP version 2.6.3  Copyright (C) 2001-2008  Dennis A. Bush"

struct uftp_h {
  int32_t uftp_id;
  int32_t func;
  uint32_t tx_id;
  int32_t seq_num;
  int32_t section_num;
  int32_t nak_count;
  int32_t blsize;
  int32_t pass;
};

/* As of 2.6.2 all code buffers for filenames are 100 bytes.
 * Leave it at 300 bytes here for backward compatibility of the protocol.
 */
struct fileinfo {
  uint32_t fsize;
  int32_t block_total;
  int32_t section_total;
  int32_t open;
  struct in_addr mcast;
  struct in_addr addr_list[MAXDEST];
  char name[300];
  int16_t rxlatency1;
  int16_t rxlatency2;
  uint32_t hilargefsize,lolargefsize;
};

char *strfunc(int func)
{
  switch (func) {
  case ANNOUNCE:
    return "ANNOUNCE";
  case REGISTER:
    return "REGISTER";
  case REG_CONF:
    return "REG_CONF";
  case FILESEG:
    return "FILESEG";
  case DONE:
    return "DONE";
  case STATUS:
    return "STATUS";
  case DONE_CONF:
    return "DONE_CONF";
  case ABORT:
    return "ABORT";
  default:
    return "UNKNOWN";
  }
}

void head_hton(struct uftp_h *header)
{
  header->uftp_id=htonl(header->uftp_id);
  header->func=htonl(header->func);
  header->tx_id=htonl(header->tx_id);
  header->seq_num=htonl(header->seq_num);
  header->section_num=htonl(header->section_num);
  header->nak_count=htonl(header->nak_count);
  header->blsize=htonl(header->blsize);
  header->pass=htonl(header->pass);
}
 
void head_ntoh(struct uftp_h *header)
{
  header->uftp_id=ntohl(header->uftp_id);
  header->func=ntohl(header->func);
  header->tx_id=ntohl(header->tx_id);
  header->seq_num=ntohl(header->seq_num);
  header->section_num=ntohl(header->section_num);
  header->nak_count=ntohl(header->nak_count);
  header->blsize=ntohl(header->blsize);
  header->pass=ntohl(header->pass);
}
 
void info_hton(struct fileinfo *info)
{
  info->fsize=htonl(info->fsize);
  info->block_total=htonl(info->block_total);
  info->section_total=htonl(info->section_total);
  info->open=htonl(info->open);
  info->rxlatency1=htons(info->rxlatency1);
  info->rxlatency2=htons(info->rxlatency2);
  info->hilargefsize=htonl(info->hilargefsize);
  info->lolargefsize=htonl(info->lolargefsize);
}
 
void info_ntoh(struct fileinfo *info)
{
  info->fsize=ntohl(info->fsize);
  info->block_total=ntohl(info->block_total);
  info->section_total=ntohl(info->section_total);
  info->open=ntohl(info->open);
  info->rxlatency1=ntohs(info->rxlatency1);
  info->rxlatency2=ntohs(info->rxlatency2);
  info->hilargefsize=ntohl(info->hilargefsize);
  info->lolargefsize=ntohl(info->lolargefsize);
}

void logfunc(FILE *logf, int32_t id, int newline, int err, int sockerr, char str[], ...)
{
  va_list args;
 
  if (id)
    fprintf(logf,"[%08lX]: ",id);
  va_start(args,str);
  vfprintf(logf,str,args);
  va_end(args);
  if (sockerr) {
#ifdef WINDOWS
    fprintf(logf,": (%d) %s",WSAGetLastError(),GetWinsockErrString(WSAGetLastError()));
#else
    fprintf(logf,": %s",strerror(err));
#endif
  } else if (err) {
    fprintf(logf,": %s",strerror(err));
  } 
  if (newline) {
    fprintf(logf,"\n");
  }
  fflush(logf);
}
